/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */

#define OverConstantA(s,d,t1)									        \
		"pmullw %%mm0, "s"\n\t"      /*   s: 0|Aa*Ra|Aa*Ga|Aa*Bb */			        \
		"pmullw %%mm3, "d"\n\t"      /*   d: 0|Rb*(255-Aa)|Gb*(255-Aa)|Bc*(255-Aa) */	        \
		Div255(s,t1)                 		                                                \
		Div255(d,t1)										\
		"paddw "s", "d"\n\t"          /*   d: 00Ac00Rc00Gc00Bc */

#define Expand565Small(m,t1,c) 	"movq "m", "t1"\n\t"		\
				"psllq $(16 - 5), "m"\n\t"	\
				"por "m", "t1"\n\t"		\
				"psllq $((36 - 11) - (16 -5)), "m"\n\t"	\
				"por "t1", "m"\n\t"		\
				"pand 8("c"), "m"\n\t"		\
				"pmullw 16("c"), "m"\n\t"       \
                                "psrlw $8, "m"\n\t"
#define COMPOSE(t1,t2)					\
	  (((t1 + (t1 >> 8)) >> 8) + ((t2 + (t2 >> 8)) >> 8))

#define COMPOSITE_0565_0565_VARS(tn)						\
  CARD32 tn##r0, tn##g0, tn##b0;						\
  CARD32 tn##r1, tn##g1, tn##b1;						\
  CARD32 tn##r2, tn##g2, tn##b2;						\
  CARD32 tn##r3, tn##g3, tn##b3;						\
  CARD32 tn##r4, tn##g4, tn##b4

#define COMPOSITE_0565_0565_PIXEL(s,d,r,tn)					\
  tn##r0 = (d & 0xf800);							\
  tn##r1 = (255 - alpha) * ((tn##r0 >> 8) + (tn##r0 >> 13)) + 0x80;		\
  tn##r2 = (s & 0xf800);							\
  tn##r3 = alpha * ((tn##r2 >> 8) + (tn##r2 >> 13)) + 0x80;			\
  tn##r4 = COMPOSE (tn##r1, tn##r3);						\
										\
  tn##g0 = (d & 0x07e0);							\
  tn##g1 = (255 - alpha) * ((tn##g0 >> 3) + (tn##g0 >> 9)) + 0x80;		\
  tn##g2 = (s & 0x07e0);							\
  tn##g3 = alpha * ((tn##g2 >> 3) + (tn##g2 >> 9)) + 0x80;			\
  tn##g4 = COMPOSE (tn##g1, tn##g3);						\
										\
  tn##b0 = (d & 0x001f);							\
  tn##b1 = (255 - alpha) * ((tn##b0 << 3) + (tn##b0 >> 2)) + 0x80;		\
  tn##b2 = (s & 0x001f);							\
  tn##b3 = alpha * ((tn##b2 << 3) + (tn##b2 >> 2)) + 0x80;			\
  tn##b4 = COMPOSE (tn##b1, tn##b3);						\
										\
  r = (((tn##r4 & 0xf8) << 8) | ((tn##g4 & 0xfc) << 3) | ((tn##b4 >> 3)))

static CARD16
composite_0565_0565_pixel (CARD32 s,
			   CARD16 d,
			   CARD8  alpha)
{
  COMPOSITE_0565_0565_VARS(t);
  CARD32 r;

  COMPOSITE_0565_0565_PIXEL(s,d,r,t);

  return r;
}

void
composite_0565_0565_mmx  (CARD8 *src_buf,
			  int     src_rowstride,
			  CARD8  alpha,
			  CARD8 *dest_buf,
			  int     dest_rowstride,
			  int     width,
			  int     height)
{
  CARD8 *src = src_buf;
  CARD8 *dest = dest_buf;
  CARD32 a = (alpha << 16) | (alpha << 8) | alpha;
  CARD32 b = -1 ^ a;
  
  MMXPrepare();
  
  __asm__ __volatile__ (Load8888 ("%1", "%%mm3")
			Load8888 ("%0", "%%mm2")
			"movq %%mm2, %%mm0"
			: /* no output */
			: "r" (a), "r" (b));
  
  while (height--)
    {
      int twidth = width;
      CARD8 *p = src;
      CARD8 *q = dest;

      if (twidth && ((unsigned long)q & 3))
	{
	  *(CARD16 *)q = composite_0565_0565_pixel(*(CARD16 *)p, *(CARD16 *)q, alpha);
	  q += 2;
	  p += 2;
	  twidth--;
	}
      
      while (twidth >= 2)
	{
	  twidth -= 2;

	  __asm__ __volatile__ ("movd %1, %%mm2\n\t"
				"movd %0, %%mm4\n\t"
				"movq %%mm2, %%mm7\n\t"
				"movq %%mm4, %%mm5\n\t"
				"pand 72(%2), %%mm2\n\t"
				"pand 72(%2), %%mm4\n\t"
				Expand565Small("%%mm2", "%%mm6","%2")
				Expand565Small("%%mm4", "%%mm6","%2")
				OverConstantA("%%mm4", "%%mm2", "%%mm6")
				Pack565("%%mm2", "%%mm6","%%mm4","%2")
				
				"psrlq $16, %%mm7\n\t"
				"psrlq $16, %%mm5\n\t"
				Expand565Small("%%mm7", "%%mm6","%2")
				Expand565Small("%%mm5", "%%mm6","%2")
				OverConstantA("%%mm5", "%%mm7", "%%mm6")
				Pack565("%%mm7", "%%mm6","%%mm4","%2")
				"psllq $16, %%mm7\n\t"

				"por %%mm7, %%mm2\n\t"
				"movd %%mm2, %1"
				
				: /* no output */
				: "m" (*p), "m" (*q), "r" (mmx_constants));
	  p += 4;
	  q += 4;
	}
      
      if (twidth)
	*(CARD16 *)q = composite_0565_0565_pixel(*(CARD16 *)p, *(CARD16 *)q, alpha);
      
      src += src_rowstride;
      dest += dest_rowstride;
    }

  __asm__ __volatile__ ("emms");
}
