/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */

void
composite_0565_0565  (CARD8 *src_buf,
		      int    src_rowstride,
		      CARD8  alpha,
		      CARD8 *dest_buf,
		      int    dest_rowstride,
		      int    width,
		      int    height)
{
  CARD8 *src = src_buf;
  CARD8 *dest = dest_buf;
  
  while (height--)
    {
      int twidth = width;
      CARD8 *p = src;
      CARD8 *q = dest;

#define TWO_PIXELS_AT_A_TIME
      
#ifdef TWO_PIXELS_AT_A_TIME
      if (twidth && ((unsigned long)q & 3))
	{
	  *(CARD16 *)q = composite_0565_0565_pixel(*(CARD16 *)p, *(CARD16 *)q, alpha);
	  q += 2;
	  p += 2;
	  twidth--;
	}

      while (twidth >= 2)
	{
	  COMPOSITE_0565_0565_VARS(ta);
	  COMPOSITE_0565_0565_VARS(tb);
	  CARD32 s = *(CARD32 *)p;
	  CARD32 d = *(CARD32 *)q;
	  CARD32 sb, db;
	  CARD32 ra, rb;

	  COMPOSITE_0565_0565_PIXEL(s,d,ra,ta);

	  sb = s >> 16;
	  db = d >> 16;
	  
	  COMPOSITE_0565_0565_PIXEL(sb,db,rb,tb);

	  *(CARD32 *)q = (rb << 16) | ra ;
	  
	  twidth -= 2;

	  p += 4;
	  q += 4;
	}
      
      if (twidth)
	*(CARD16 *)q = composite_0565_0565_pixel(*(CARD16 *)p, *(CARD16 *)q, alpha);
#else /* !TWO_PIXELS_AT_A_TIME */
      while (twidth--)
	{
	  COMPOSITE_0565_0565_VARS(t);
	  CARD16 s = *(CARD16 *)p;
	  CARD16 d = *(CARD16 *)q;
	  CARD16 r;
	  
	  COMPOSITE_0565_0565_PIXEL(s,d,r,t);
	  
	  *(CARD16 *)q = r;
	  
	  p += 2;
	  q += 2;
	}
      
      if (twidth)
	*(CARD16 *)q = composite_0565_0565_pixel(*(CARD16 *)p, *(CARD16 *)q, alpha);
#endif
      
      src += src_rowstride;
      dest += dest_rowstride;
    }
}

#undef COMPOSE
