/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */

#ifdef PREMULT
#define COMPOSE(C,t1,t2)			\
	  (t2 = p[C] + ((t1 + (t1 >> 8)) >> 8),	\
	   (t2 | (0 - (t2 >> 8))))
#else
#define COMPOSE(C,t1,t2)					\
	  (t2 = p[A] * p[C] + 0x80,				\
           ((t1 + (t1 >> 8)) >> 8) + ((t2 + (t2 >> 8)) >> 8))
#endif

void
NAME (CARD8 *src_buf,
      int    src_rowstride,
      CARD8 *dest_buf,
      int    dest_rowstride,
      int    width,
      int    height)
{
  CARD8 *src = src_buf;
  CARD8 *dest = dest_buf;

  while (height--)
    {
      int twidth = width;
      CARD8 *p = src;
      CARD32 *q = (CARD32 *)dest;

      if (twidth && ((unsigned long)q & 3))
	{
	  *(CARD16 *)q = PIXEL(*(CARD32 *)p, *(CARD16 *)q);
	  q = (CARD32 *)(1 + (CARD16 *)q);
	  p += 4;
	  twidth--;
	}
      
      while (twidth >= 2)
	{
	  CARD32 tr0, tg0, tb0;
	  CARD32 tr1, tg1, tb1;
	  CARD32 tr2, tg2, tb2;
	  CARD32 tr3, tg3, tb3;
	  CARD32 tr4, tg4, tb4;
	  CARD32 tr5, tg5, tb5;
	  CARD32 tr6, tg6, tb6;
	  CARD32 tr7, tg7, tb7;
	  CARD32 t = *q;

	  twidth -= 2;

#if G_BYTE_ORDER == G_BIG_ENDIAN	  
	  tr4 = (t & 0xf8000000);
	  tr5 = (255 - p[A]) * ((tr4 >> (8 + 16)) + (tr4 >> (13 + 16))) + 0x80;
	  tr7 = COMPOSE(R,tr5,tr6);

	  tg4 = (t & 0x07e00000);
	  tg5 = (255 - p[A]) * ((tg4 >> (3 + 16)) + (tg4 >> (9 + 16))) + 0x80;
	  tg7 = COMPOSE(G,tg5,tg6);
	  
	  tb4 = (t & 0x001f0000);
	  tb5 = (255 - p[A]) * ((tb4 >> (16 - 3)) + (tb4 >> (2 + 16))) + 0x80;
	  tb7 = COMPOSE(B,tb5,tb6);

	  p += 4;
	  tr0 = (t & 0xf800);
	  tr1 = (255 - p[A]) * ((tr0 >> 8) + (tr0 >> 13)) + 0x80;
	  tr3 = COMPOSE(R,tr1,tr2);
	  
	  tg0 = (t & 0x07e0);
	  tg1 = (255 - p[A]) * ((tg0 >> 3) + (tg0 >> 9)) + 0x80;
	  tg3 = COMPOSE(G,tg1,tg2);
	  
	  tb0 = (t & 0x001f);
	  tb1 = (255 - p[A]) * ((tb0 << 3) + (tb0 >> 2)) + 0x80;
	  tb3 = COMPOSE(B,tb1,tb2);
#else
	  tr0 = (t & 0xf800);
	  tr1 = (255 - p[A]) * ((tr0 >> 8) + (tr0 >> 13)) + 0x80;
	  tr3 = COMPOSE(R,tr1,tr2);
	  
	  tg0 = (t & 0x07e0);
	  tg1 = (255 - p[A]) * ((tg0 >> 3) + (tg0 >> 9)) + 0x80;
	  tg3 = COMPOSE(G,tg1,tg2);
	  
	  tb0 = (t & 0x001f);
	  tb1 = (255 - p[A]) * ((tb0 << 3) + (tb0 >> 2)) + 0x80;
	  tb3 = COMPOSE(B,tb1,tb2);

	  p += 4;
	  tr4 = (t & 0xf8000000);
	  tr5 = (255 - p[A]) * ((tr4 >> (8 + 16)) + (tr4 >> (13 + 16))) + 0x80;
	  tr7 = COMPOSE(R,tr5,tr6);

	  tg4 = (t & 0x07e00000);
	  tg5 = (255 - p[A]) * ((tg4 >> (3 + 16)) + (tg4 >> (9 + 16))) + 0x80;
	  tg7 = COMPOSE(G,tg5,tg6);
	  
	  tb4 = (t & 0x001f0000);
	  tb5 = (255 - p[A]) * ((tb4 >> (16 - 3)) + (tb4 >> (2 + 16))) + 0x80;
	  tb7 = COMPOSE(B,tb5,tb6);
#endif	  
	  
	  *q = (((tr3 & 0xf8) << 8)      | ((tg3 & 0xfc) << 3)         | ((tb3 >> 3)) |
		(tr7 & 0xf8) << (8 + 16) | ((tg7 & 0xfc) << (3 + 16) ) | ((tb7 & 0xf8) << (16 - 3)));

	  p += 4;
	  q++;
	}

      if (twidth)
	*(CARD16 *)q = PIXEL(*(CARD32 *)p, *(CARD16 *)q);
      
      src += src_rowstride;
      dest += dest_rowstride;
    }
}

#undef COMPOSE
