/* Copyright © 2002 Red Hat, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of Red Hat not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  Red Hat makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * RED HAT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL RED HAT
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN 
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 * Author:  Owen Taylor <otaylor@redhat.com>
 */

void
NAME (CARD8 *src_buf,
      int    src_rowstride,
      CARD8 *dest_buf,
      int    dest_rowstride,
      int    width,
      int    height)
{
  CARD8 *src = src_buf;
  CARD8 *dest = dest_buf;

  while (height--)
    {
      int twidth = width;
      CARD8 *p = src;
      CARD32 *q = (CARD32 *)dest;

      while (twidth--)
	{
	  CARD32 ta1, tr1, tg1, tb1;
	  CARD32 ta2, tr2, tg2, tb2;
	  CARD32 tr3, tg3, tb3;
	  CARD32 t = *q;

#ifdef PREMULT
	  tr1 = (255 - p[A]) * FbGet8(t,16) + 0x80;
	  tr2 = ((tr1 + (tr1 >> 8)) >> 8) + p[R];
	  tr3 = tr2 | (0 - (tr2 >> 8));

	  tg1 = (255 - p[A]) * FbGet8(t,8) + 0x80;
	  tg2 = ((tg1 + (tg1 >> 8)) >> 8) + p[G];
	  tg3 = tg2 | (0 - (tg2 >> 8));

	  tb1 = (255 - p[A]) * FbGet8(t,0) + 0x80;
	  tb2 = ((tb1 + (tb1 >> 8)) >> 8) + p[B];
	  tb3 = tb2 | (0 - (tb2 >> 8));
#else	  
	  tr1 = (255 - p[A]) * FbGet8(t,16) + 0x80;
	  tr2 = p[A] * p[R] + 0x80;
	  tr3 = ((tr1 + (tr1 >> 8)) >> 8) + ((tr2 + (tr2 >> 8)) >> 8);

	  tg1 = (255 - p[A]) * FbGet8(t,8) + 0x80;
	  tg2 = p[A] * p[G] + 0x80;
	  tg3 = ((tg1 + (tg1 >> 8)) >> 8) + ((tg2 + (tg2 >> 8)) >> 8);

	  tb1 = (255 - p[A]) * FbGet8(t,0) + 0x80;
	  tb2 = p[A] * p[B] + 0x80;
	  tb3 = ((tb1 + (tb1 >> 8)) >> 8) + ((tb2 + (tb2 >> 8)) >> 8);
#endif
	  ta1 = (255 - p[A]) * FbGet8(t,24) + 0x80;
	  ta2 = ((ta1 + (ta1 >> 8)) >> 8) + p[A];

	  *q = (ta2 << 24) | (tr3 << 16) | (tg3 << 8) | tb3;

	  p += 4;
	  q++;
	}
      
      src += src_rowstride;
      dest += dest_rowstride;
    }
}
